;;-----------------------------LICENSE NOTICE------------------------------------
;;  This file is part of BringRGBack: An Amstrad CPC Game 
;;  Copyright (C) 2018 Ivan Palacios Verano / Edward Thomas Owen Nolla / Javier Richarte Gras
;;
;;  BringRGBack is free software: you can redistribute it and/or modify
;;  it under the terms of the GNU Lesser General Public License as published by
;;  the Free Software Foundation, either version 3 of the License, or
;;  (at your option) any later version.
;;
;;  BringRGBack is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU Lesser General Public License for more details.
;;
;;  You should have received a copy of the GNU Lesser General Public License
;;  along with BringRGBack.  If not, see <http://www.gnu.org/licenses/>.
;;-------------------------------------------------------------------------------

.include "bullet.h.s"
.include "main.h.s"
.include "render.h.s"
.include "enemy.h.s"
.include "obstacle.h.s"
.include "powerUp.h.s"
.include "collision.h.s"
.include "entity.h.s"
.include "player.h.s"
.include "sound.h.s"

;;LOCAL VARIABLES
m_next_bullet: .dw bullet_vector       ;;Pointer of the memory position of the next entity 
m_num_bullet: .db 00                   ;;Counter to see how many entities we have created
bullet_mem: .dw 0x0000
m_bullet_velocity: .dw #0x0000
m_powerUp_act: .db #0
m_powerUpCounter: .db #0

;; == Auxiliar Variables == ;;
m_bullet_counter: .db #0x00

;;============================== CONSTANTS =======================================;;

;;== Bullets ==;;
k_max_num_bullets = 10                    ;;Constant for the maximum number of bullets
k_bullet_size = bullet_vector0_size       ;;Dynamic size of Bullet
m_bullet_lft = 15                         ;;Life cycles of a bullet

;;== Bullet vector ==;;
bullet_vector:
DefineNBullets bullet_vector, k_max_num_bullets

;;== Default bullet for new instances, i.e. the default constructor ==;;
DefineBullet bullet_data, 0x00, 0x00, 0x00, 0x00, 0x01, 0x04, _bulletSprite, 0x00, 0x00, 0x00, 0x00, bullet_CheckLFT, m_bullet_lft, 0, #1
DefineBullet laser_data, 0x00, 0x00, 0x00, 0x00, 0x02, 0x08, _laserSprite, 0x00, 0x00, 0x00, 0x00, bullet_CheckLFT, m_bullet_lft, 0, #2

;;==============================================================================
;; BULLET_NEW: Register a new bullet
;;      INPUT: 
;;      DESTROYS: AF, BC, HL 
;;      OUTPUT: 
;;          HL -> Pointer to the current bullet that we are going to create
;;==============================================================================

bullet_New:

    call bullet_CheckPUActivation  ;;Decide the type of bullets to shot

    ld  a, (m_num_bullet)
    inc a                          ;;A++
    ld (m_num_bullet), a           ;;<>

    ld hl, (m_next_bullet)         ;;HL = Pointer to the last active bullet    
    ld bc, #k_bullet_size          ;;BC = Bullet's size
    add hl, bc                     ;;HL += BC -> the position of the next new bullet
    ld (m_next_bullet), hl         ;;NEXT_ENTITY = HL (last active bullet)
    or  a                          ;;To put Carry to 0 (a or a = 0)
    sbc hl, bc                     ;;HL -= BC -> to get the pointer to the current entity

    ex de, hl

    ld  a, (m_powerUp_act)         ;;If PowerUp == Activated (1)
    cp  #0                         ;;|
    jr  nz, new_laser              ;;|

        ld  hl, #bullet_data       ;;Else -> Create a new bullet
        ret

    new_laser:                     ;;\ Create a new laser
    ld  hl, #laser_data

ret



;;============================================================
;; BULLET_COPY: Copy the values of one bullet to another 
;; INPUT:
;;      HL -> Origin data to copy
;; DESTROY: 
;;      BC, IY
;; OUTPUT:
;;      IY -> Pointer to the new bullet/laser created
;;============================================================

bullet_Copy:

    ld (bullet_mem), de
    ld iy, (bullet_mem)            ;;IY = Pointer to the new bullet created
    ld  bc, #k_bullet_size         ;;Copy the size of the entity
    ldir 

ret

;;============================================================
;; BULLET_DESTROY: Deletes a bullet from the memory
;; INPUT:   
;;    IX -> Pointer to THIS bullet
;; DESTROYS:
;;    AF, HL, DE, BC
;;============================================================

bullet_Destroy:
    call ren_Clear_FrontBack
    
    ld  a, (m_num_bullet)
    cp  #0
    jr  z, exit_destroy

        call bullet_TurnBackNextBullet
        call bullet_DecreaseNumBullets

        push ix
        pop de   

        ld  a, h                                ;;Check if the last bullet is the same as THIS
        cp  d                                   ;;|
        jr nz, make_copy                        ;;|

        ld  a, l                                ;;|
        cp  e                                   ;;|
        jr nz, make_copy                        ;;|

        call bullet_SetDeadState
        ret	

        make_copy:
        ld  bc, #k_bullet_size                  ;;Copy the last bullet of the vector
        ldir                                    ;;to the position of THIS bullet that we want to destroy
        call bullet_SetDeadState
        
exit_destroy:
ret

;;=======================================================================
;;BULLET_TURNBACKNEXTBULLET: Set back next bullet pointer
;; DETROY: AF, HL, BC 
;;=======================================================================

bullet_TurnBackNextBullet:

    ld  hl, (m_next_bullet)                 ;;HL = Pointer to the last active bullet                                                          
    ld  bc, #k_bullet_size                  
    or a 
    sbc hl, bc
    ld  (m_next_bullet), hl  

ret

;;=======================================================================
;;BULLET_DECREASENUMBULLETS: Decrease the number of bullets 
;; DETROY: AF 
;;=======================================================================

bullet_DecreaseNumBullets:

    ld  a, (m_num_bullet)                   ;;A = number of active bullets
    dec a                                   ;;A --
    ld  (m_num_bullet), a                   ;;<>
ret

;;=======================================================================
;;BULLET_SETDEADSTATE: Set the last bullet's dead state to 1
;; DETROY: A
;;=======================================================================

bullet_SetDeadState:
    push ix

        ld ix, (m_next_bullet)
        ld a, #1
        ld b_dead(ix), a
        
    pop ix 
ret

;;=======================================================================
;; bullet_CHECKPUSACTIVATION: Checks if the player has burnt-out the PU 
;; INPUT:
;;      IX -> Pointer to bullet
;; DESTROYS:
;;      AF, HL
;;=======================================================================

bullet_CheckPUActivation:

    ld hl, #m_powerUp_act               ;;If PowerUp is not activated
    ld  a, (hl)                         ;;|
    cp  #1                              ;;|
    jr  nz, exit_checking               ;;\ Exit

        ld hl, #m_powerUpCounter        ;;In other case: check how many Laser bullets you have
        ld  a, (hl)                     ;;to shot
        dec a                           ;;|
        jp  z, bullet_DesactivatePU     ;;\If lasers == 0 -> Desactivate PowerUp

        ld  (hl), a                     ;;PowerUp counter --

exit_checking:        
ret

;;=======================================================================
;; BULLET_DOFORALL: Update, draw and clear to all bullets
;; INPUT:  HL-> Function 
;;=======================================================================

bullet_DoForAll:

    ld a, (m_num_bullet)           ;;Loads the number of the entities 
    cp #0
    jr z, not_dfa  

    ld ix, #bullet_vector          ;;Loads the first entity of the vector on IX
    ld (method), hl                ;;HL = function parameter
    
    loop:

        push af                    ;;Push a on the stack to not lose it

        ld a, b_dead(ix)
        cp #1
        jr z, dead

        method = . + 1             ;;Current position + 1 = the function of the next call
        call bullet_CheckLFT       ;;We call the method we have in hl 

        dead:

        pop af

        ld bc, #k_bullet_size      ;;Move M_NEXT_BULLET pointer to the next position by the bullets size
        add ix, bc              
        dec a            
    jr nz, loop                    ;;Next iteration

not_dfa:
ret


;;===========================================================
;; BULLET_CHECKLFT (*): 
;;
;; (*) -> Tail recursion
;;===========================================================

bullet_CheckLFT:

    ld hl, #obstacle_BulletCollisions
    call obstacle_EveryObstacle
    call enemy_EveryEnemy
    call bullet_CheckScreenCollissions 

    ld   a, b_lft(ix)               ;;Check the lifetime time of ix bullet
    cp   #0                         ;;|
    jr   z, bullet_die              ;;\ If it's time to die, then stop updating

        dec a   
        ld  b_lft(ix), a
        call bullet_Move
        ret
        
    bullet_die:
    jp bullet_Destroy

;;==================================================================================
;; BULLET_CHECKCOLISSIONS: Check the bullets colissions with the screen's limits
;; INPUT:
;;      IX -> Pointer to Bullet
;; DESTROYS:
;;      AF, BC
;;=================================================================================

bullet_CheckScreenCollissions:

    ;;
    ;;Check if the bullet is going out by the left side screen
    ;;
    ld  a, de_x(ix)      ;;A = B_X (X position of the bullet)
    cp  #17            ;;A -= 1
    jr  z, die_b        ;;If z = 1, then destroy bullet
    jp  m, die_b        ;;OR If negative result of CP

    ld  a, #80-2        ;;A = 78 (right side screen limit)       
    ld  b, de_x(ix)      ;;B = B_X (X position of the bullet)
    cp  b               ;;A -= B
    jr  z, die_b        ;;If z = 1, then destroy the bullet
    jp  m, die_b        ;;OR negative result of CP

    ld  a, de_y(ix)      ;;A = B_X (X position of the bullet)
    cp  #0+5            ;;A -= 5
    jr  z, die_b        ;;If z = 1, then destroy bullet
    jp  c, die_b        ;;We have to worry about the Carry in this case because if we do 90 - 5

    ld  a, #200-10      ;;A = 190 (down side screen limit)       
    ld  b, de_y(ix)      ;;B = B_X (X position of the bullet)
    cp  b               ;;A -= B
    jr  z, die_b        ;;If z = 1, then destroy the bullet
    jp  c, die_b        ;;We have to worry about the Carry in this case because if we do 90 - 5

    ret                 ;;In other case, don't destroy the bullet
    
    die_b:
    ld  b_lft(ix), #0   ;;Destroy the bullet

ret

;;=========================================================================================
;; ENEMY_CHECKCOLLISION: Check the collision between the array of enemies and the hero
;; INPUT: IX -> BULLET
;;        IY -> ENEMY
;;=========================================================================================

bullet_Collissions:
    
    call collision_CheckEntitiesCollision

    cp #0                ;;NO COLLISION
    jr z, no_collision

    ;;COLLISION
    ld  d, b_damage(ix)                 ;;B = bullet's damage
    call enemy_Hit                      ;;Hit enemy with B damage
    pop af                              ;;
    pop af                              ;;
    jp  bullet_die 
    
    no_collision:
ret

;;===========================================================
;; BULLET_MOVE: Moves the bullet by it velocity
;; INPUT:
;;      IX -> Pointer to Bullet
;; DESTROYS:
;;      AF
;;===========================================================

bullet_Move:

   ld    a, de_x(ix)        ;;B_X += B_X + B_VX
   add   b_vx(ix)          ;;|
   ld    de_x(ix), a        ;;\

   ld    a, de_y(ix)        ;;B_Y += B_Y + B_VY
   add   b_vy(ix)          ;;|
   ld    de_y(ix), a        ;;\

ret

;;===========================================================
;; BULLET_GETVELOCITY: Returns the bullet velocity
;; OUTPUT:
;;      HL -> Pointer to m_bullet_velocity
;; DESTROYS: 
;;      HL
;;===========================================================

bullet_GetVelocity:

    ld hl, #m_bullet_velocity

ret

;;===========================================================
;; BULLET_GETNUMBULLET: Returns the number of active bullets
;; DESTROYS: 
;;      AF
;;===========================================================

bullet_GetNumBullet:

    ld  a, (m_num_bullet)

ret

;;=============================================================
;; BULLET_GETBULLETSIZE: Returns the maximum number of bullets
;; DESTROYS: 
;;      AF
;;=============================================================

bullet_GetBulletsSize:

    ld  a, (k_max_num_bullets)

ret

;;=====================================================================
;; bullet_DestroyEveryBullet: destroy every bullet (captain obvious)
;; DESTROY: HL, A
;;=====================================================================

bullet_DestroyEveryBullet:

    call bullet_DesactivatePU
    push ix
    ld hl, #bullet_Destroy
    call bullet_DoForAll
    pop ix

    ld a, #00
    ld (m_num_bullet), a        ;;Set number of bullets = 0
    ld hl, #bullet_vector      
    ld (m_next_bullet), hl        ;;Set next entity to the array's first position

ret

;;=====================================================
;; bullet_ActivatePU: activates the Power Up 
;; DESTROYS:
;;        HL
;;=====================================================

bullet_ActivatePU:

    ld  hl, #m_powerUp_act
    ld  (hl), #1

    ld  hl, #m_powerUpCounter
    ld  (hl), #8

    ld  a, #4
    ld  (m_fire_delay), a

ret

;;=====================================================
;; bullet_DesactivatePU: desactivates the Power Up 
;; DESTROYS:
;;        HL
;;=====================================================

bullet_DesactivatePU:

    ld  hl, #m_powerUp_act
    ld  (hl), #0

    ld  hl, #m_powerUpCounter
    ld  (hl), #0

    ld  a, #3
    ld  (m_fire_delay), a

ret








